"""
motor_controller controller

A simple controller to set the target of a robot
"""

import csv
import json
import os
from controller import Robot
import numpy as np
from scipy.spatial.transform import Rotation

import robots

# Setting parameters from environment variables
robot_name = os.environ.get('ROBOT_NAME', 'RobotRRR')
robot_targets = {
    'RobotRT': [1.57, 0.1],
    'RobotRRR': [1.57, 1.57, 1.57]
}
targets = robot_targets[robot_name]

model = robots.getRobotModel(robot_name)
robot = Robot()
timestep = int(robot.getBasicTimeStep())

# Initializing motors
motors = []
for name in model.getMotorsNames():
    motors.append(robot.getDevice(name))

# Initializing sensors
sensors = []
for name in model.getSensorsNames():
    sensors.append(robot.getDevice(name))
    sensors[-1].enable(timestep)

with open('robot_data.csv', 'w') as output_file:
    logger = csv.DictWriter(output_file, ["t", "variable", "order", "source", "value"])
    logger.writeheader()

    t = 0.0  # [s]
    while robot.step(timestep) != -1:
        nb_joints = model.getNbJoints()
        q = np.zeros(nb_joints)
        for i in range(nb_joints):
            motors[i].setPosition(targets[i])
            q[i] = sensors[i].getValue()
            joint_name = model.getJointsNames()[i]
            logger.writerow({"t": t, "variable": joint_name, "order": 0, "source": "target", "value": targets[i]})
            logger.writerow({"t": t, "variable": joint_name, "order": 0, "source": "sensor", "value": q[i]})
        op_pos = model.computeMGD(q)
        op_dim_names = model.getOperationalDimensionNames()
        base_from_tool = model.getBaseFromToolTransform(q)
        tool_pos = base_from_tool[:3, 3]
        tool_quat = Rotation.from_matrix(base_from_tool[:3, :3]).as_quat()
        custom_data = json.dumps({'position': tool_pos.tolist(), 'orientation': tool_quat.tolist()})
        robot.setCustomData(custom_data)
        for name, idx in {"x": 0, "y": 1, "z": 2}.items():
            logger.writerow({"t": t, "variable": name, "order": 0, "source": "controller", "value": tool_pos[idx]})
        for name, idx in {"quat_x": 0, "quat_y": 1, "quat_z": 2, "quat_w": 3}.items():
            logger.writerow({"t": t, "variable": name, "order": 0, "source": "controller", "value": tool_quat[idx]})
        t += timestep / 10**3
