import numpy as np
from abc import abstractmethod
import homogeneous_transform as ht


class RobotModel:
    def getNbJoints(self):
        """
        Returns
        -------
        length : int
            The number of joints for the robot
        """
        return len(self.getJointsNames())

    def getMotorsNames(self):
        """
        Returns
        -------
        motor_names : string[]
            The list of names for the motors
        """
        return [name + "_motor" for name in self.getJointsNames()]

    def getSensorsNames(self):
        """
        Returns
        -------
        motor_names : string[]
            The list of names for the motors
        """
        return [name + "_sensor" for name in self.getJointsNames()]

    @abstractmethod
    def getJointsNames(self):
        """
        Returns
        -------
        joint_names : string[]
            The list of names given to robot joints
        """

    @abstractmethod
    def getJointsLimits(self):
        """
        Returns
        -------
        np.array
            The values limits for the robot joints, each row is a different
            joint, column 0 is min, column 1 is max
        """

    @abstractmethod
    def getOperationalDimensionNames(self):
        """
        Returns
        -------
        joint_names : string array
            The list of names of the operational dimensions
        """

    @abstractmethod
    def getBaseFromToolTransform(self, joints):
        """
        Parameters
        ----------
        joints_position : np.array
            The values of the joints of the robot in joint space

        Returns
        -------
        np.array
            The transformation matrix from base to tool
        """

    @abstractmethod
    def computeMGD(self, q):
        """
        Parameters
        ----------
        q : np.array
            The values of the joints of the robot in joint space

        Returns
        -------
        np.array
            The coordinate of the effectors in the operational space
        """


class RobotRT(RobotModel):
    """
    Model a robot with a 2 degrees of freedom: 1 rotation and 1 translation

    The operational space of the robot is 2 dimensional because it can only move inside a plane
    """
    def __init__(self):
        W = 0.05
        L0 = 1.0
        L1_offset = 0.2
        L2 = 0.25
        self.T_0_1 = ht.translation([0, 0, L0+W/2])
        self.T_1_2 = ht.translation([L1_offset, 0, 0])
        self.T_2_E = ht.translation([0.0, -(W/2 + L2), 0]) @ ht.rot_z(np.pi)

    def getJointsNames(self):
        return ["q1", "q2"]

    def getJointsLimits(self):
        return np.array([[-np.pi, np.pi], [0, 0.25]], dtype=np.double)

    def getOperationalDimensionNames(self):
        return ["x", "y"]

    def getBaseFromToolTransform(self, joints):
        T_0_1 = self.T_0_1 @ ht.rot_z(joints[0])
        T_1_2 = self.T_1_2 @ ht.translation(joints[1] * np.array([1, 0, 0]))
        return T_0_1 @ T_1_2 @ self.T_2_E

    def computeMGD(self, q):
        tool_pos = self.getBaseFromToolTransform(q) @ np.array([0, 0, 0, 1])
        return tool_pos[:2]


class RobotRRR(RobotModel):
    """
    Model a robot with 3 degrees of freedom along different axis
    """
    def __init__(self):
        W = 0.05
        L0 = 1.0
        L1 = 0.5
        L2 = 0.4
        L3 = 0.3
        self.T_0_1 = ht.translation([0, 0, L0+W/2])
        self.T_1_2 = ht.translation([0, L1, 0])
        self.T_2_3 = ht.translation([0.0, L2, 0])
        self.T_3_E = ht.translation([0.0, L3+W/2, 0])

    def getJointsNames(self):
        return ["q1", "q2", "q3"]

    def getJointsLimits(self):
        return np.array([[-np.pi, np.pi], [-np.pi, np.pi], [-np.pi, np.pi]], dtype=np.double)

    def getOperationalDimensionNames(self):
        return ["x", "y", "z"]

    def getBaseFromToolTransform(self, joints):
        T_0_1 = self.T_0_1 @ ht.rot_z(joints[0])
        T_1_2 = self.T_1_2 @ ht.rot_x(joints[1])
        T_2_3 = self.T_2_3 @ ht.rot_x(joints[2])
        return T_0_1 @ T_1_2 @ T_2_3 @ self.T_3_E

    def computeMGD(self, q):
        tool_pos = self.getBaseFromToolTransform(q) @ np.array([0, 0, 0, 1])
        return tool_pos[:3]


def getRobotModel(robot_name):
    robot = None
    if robot_name == "RobotRT":
        robot = RobotRT()
    elif robot_name == "RobotRRR":
        robot = RobotRRR()
    else:
        raise RuntimeError("Unknown robot name: '" + robot_name + "'")
    return robot
